from datetime import datetime, timedelta
from typing import Optional

# Program imports
from ..enums.apc_status import ApcStatus
from ..enums.apc_category import ApcCategory
from ..utils import get_apc_description
from ...ubs.dtos.analysis.results import UbsAnalysisResult
from ...utils.date_util import (
    utc_now,
    utc_datetime_to_epoch_time,
    epoch_time_to_utc_datetime,
)


def safe_decode_category(x):
    try:
        return ApcCategory(x)
    except ValueError:
        try:
            return ApcCategory[x]
        except KeyError:
            return x


class ApcQuery(UbsAnalysisResult):
    decode_names = {
        "cat": "category",
        "det_name": "signature",
        "upload_to": "upload_url",
    }

    encode_names = {
        "category": "cat",
        "signature": "det_name",
        "upload_url": "upload_to",
    }

    def __init__(
        self,
        sha256=None,
        category=None,  # Assumes an ApcCategory enum, but may also be an int
        ttl: int = 0,  # seconds
        signature=None,
        known=False,
        status: ApcStatus = None,
        upload_url=None,
        query_date: Optional[datetime] = None,
    ):
        super().__init__(sha256=sha256)
        # These are manually set via overriding __setattr__
        self.category_description = None
        self.known_category = False

        self.category = category
        self.ttl = (
            ttl if ttl is not None else 0
        )  # relative (seconds) - as returned by Avira.
        self.signature = signature
        self.known = known
        self.status = status
        self.upload_url = upload_url
        self.query_date = query_date if query_date is not None else utc_now()

        super().decode_settings.update(
            {
                "category": safe_decode_category,
                "status": lambda x: ApcStatus[x],
                "query_date": lambda x: utc_now()
                if x is None
                else epoch_time_to_utc_datetime(x),
            }
        )

        super().encode_settings.update(
            {
                "category": lambda x: x.name if isinstance(x, ApcCategory) else x,
                "status": lambda x: x.name,
                "query_date": lambda x: utc_datetime_to_epoch_time(utc_now())
                if x is None
                else utc_datetime_to_epoch_time(x),
            }
        )

    def __setattr__(self, key, value):
        super(ApcQuery, self).__setattr__(key, value)
        if key == "category":
            if isinstance(value, ApcCategory):
                self.known_category = True
                self.category_description = get_apc_description(value)
            else:
                self.known_category = False
                self.category_description = None

    @property
    def ttl_absolute(self):
        return self.query_date + timedelta(seconds=self.ttl)

    @classmethod
    def from_json_object(cls, obj, *args, **kwargs):
        ret = []
        if isinstance(obj["sha256"], dict):
            for sha256, items in obj["sha256"].items():
                items["sha256"] = sha256
                ret.append(cls.from_dict(items, encoded=True))
            return ret
        else:
            return super().from_json_object(obj, args, kwargs)
