import json
import os
from enum import Enum

import requests
from flask import request, g, jsonify
from flask_principal import Permission, Need, Principal, Identity
from requests import HTTPError

from tic.shared.service import TICService, AuthException, BasePrincipal
from tic.reputation.shared.dtos.error_responses import (
    ReputationErrorResponse,
    ReputationErrorCode,
)

read_reputation_data_permission = Permission(Need("read", "reputation_data"))


class AuthHeaders(Enum):
    X_User_Context = "X-User-Context"
    X_Request_Context = "X-Request-Context"


class IdentityPrincipalData:
    def __init__(self):
        self.org_key = None
        self.request_context = None
        self.user_content = None
        self.rbac_key = None


class IdentityPrincipal(BasePrincipal):
    def __init__(self):
        self.principal = None

    def create_principal(self, app):
        self.principal = Principal(app, use_sessions=False)
        self.principal.identity_loader(self.authenticate)
        return self.principal

    def authenticate(self):
        headers = request.headers
        user_context, org_key, request_context, rbac_key = None, None, None, None
        if headers and headers.get(AuthHeaders.X_User_Context.value):
            user_context = json.loads(headers.get(AuthHeaders.X_User_Context.value))
            org_key = user_context.get("target_org").get("org_key")
            rbac_key = user_context.get("target_org").get("rbac_key")
        if headers and headers.get(AuthHeaders.X_Request_Context.value):
            request_context = json.loads(
                headers.get(AuthHeaders.X_Request_Context.value)
            )
        identity = Identity(org_key)
        self.principal.set_identity(identity)

        principal = IdentityPrincipalData()
        principal.request_context = request_context
        principal.user_context = user_context
        principal.org_key = org_key
        principal.rbac_key = rbac_key
        g.identity_principal_data = principal


class AuthClient(object):
    def __init__(self, logger):
        self.auth_session = requests.Session()
        self.logger = logger

    def authorize(
        self,
        org_key: str,
        request_context: dict,
        user_context: dict,
        headers: dict,
        permissions: dict,
    ) -> None:
        if not (request_context and user_context and org_key):
            raise AuthException("Missing authentication headers")
        url = os.environ.get("INTERNAL_INGRESS_URL") or request_context.get(
            os.environ.get("AUTH_GATEWAY_HEADER", "th_gateway_url")
        )
        if not url:
            raise AttributeError("Unable to determine auth gateway URL")

        # If url doesn't define a protocol use https
        if not url.startswith("https://") and not url.startswith("http://"):
            url = "https://" + url

        demand_url = f"{url}/access/v1/demand"

        payload = {
            "permissions": permissions,
            "entities": {"org": org_key},
            "caller": user_context["login_org"],
        }

        self.logger.debug(
            "RBAC request url: %s, payload: %s, headers: %s",
            demand_url,
            payload,
            headers,
        )

        # Check if request is Authorized
        # response code == 200 --> authorized
        # response code == 403 --> unauthorized
        response = self.auth_session.post(url=demand_url, json=payload, headers=headers)
        try:
            response.raise_for_status()
        except HTTPError as error:
            raise AuthException(str(error))


class TICIdentityService(TICService):
    def __init__(self, config, principal=None):
        super().__init__(config, principal)
        self.auth_client = AuthClient(config.logger)

    @property
    def identity_principal(self) -> IdentityPrincipalData:
        return getattr(g, "identity_principal_data", None)


def auth_identity_validate(action: str, permission: str):
    def validate(func):
        def inner_function(self, *args, **kwargs):
            if is_probe_endpoint(request.endpoint) or self.config.allow_anonymous:
                return func(self, *args, **kwargs)
            principal = self.identity_principal
            if principal:
                auth_header = dict()
                auth_header["Content-Type"] = "application/json"
                auth_header["Accept"] = "application/json"
                self.auth_client.authorize(
                    org_key=principal.org_key,
                    request_context=principal.request_context,
                    user_context=principal.user_context,
                    headers=auth_header,
                    permissions=[{"name": permission, "actions": [action]}],
                )
            # Below block just to support mcs1
            else:
                if not (self.tic_principal and self.tic_principal.customer_token):
                    return (
                        jsonify(
                            ReputationErrorResponse(
                                error_code=ReputationErrorCode.UNAUTHORIZED,
                                message="User Unauthorized",
                            ).to_dict(True, True, value_ignorelist={None})
                        ),
                        401,
                    )
                if not read_reputation_data_permission.can():
                    return (
                        jsonify(
                            ReputationErrorResponse(
                                error_code=ReputationErrorCode.FORBIDDEN,
                                message="Permission Denied",
                            ).to_dict(True, True, value_ignorelist={None})
                        ),
                        403,
                    )
            return func(self, *args, **kwargs)

        return inner_function

    return validate


def is_probe_endpoint(endpoint) -> bool:
    return endpoint in ["health_check", "ready_check", ""]
