import base64
import datetime
from enum import Enum
import hashlib
import json
from uuid import uuid4

from botocore.config import Config
from botocore.exceptions import ClientError


class S3StorageClass(Enum):
    STANDARD = 0
    REDUCED_REDUNDANCY = 1
    STANDARD_IA = 2
    ONEZONE_IA = 3
    GLACIER = 4
    DEEP_ARCHIVE = 5
    INTELLIGENT_TIERING = 6


class S3Client(object):
    def __init__(self, boto_session, boto_config=None, endpoint_url=None):
        """

        @param boto_session: pre-existing boto session
        @param boto_config: optional boto config for the s3 client
        @param endpoint_url: optional endpoint for the s3 client used for localstack
        """

        default_config = Config(signature_version="s3v4")
        if boto_config is None:
            boto_config = default_config
        else:
            # default config values will take precedence if non-default.
            # s3 version defaults to v2.
            # This will not work in regions created past Jan. 2014:
            # http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-authenticating-requests.html
            boto_config = default_config.merge(boto_config)

        self.s3_client = boto_session.client(
            "s3", config=boto_config, endpoint_url=endpoint_url
        )
        self.s3_resource = boto_session.resource(
            "s3", config=boto_config, endpoint_url=endpoint_url
        )

        self.list_objects_enable = True

    def generate_presigned_url(self, client_method, params):
        """
        Generate a presigned url given a client, its method, and arguments

        @param client_method: The client method to presign for
        @param params: The parameters normally passed to ClientMethod
        """
        self.s3_client.generate_presigned_url(ClientMethod=client_method, Params=params)

    def upload_file(self, file_name, bucket_name, key_name, config=None):
        """
        Upload a given file to a specified s3 bucket

        @param config:
        @param file_name:
        @param bucket_name: target s3 bucket
        @param key_name: The name of the key to upload to
        """
        args = {"ServerSideEncryption": "AES256"}
        self.s3_client.upload_file(
            Filename=file_name,
            Bucket=bucket_name,
            Key=key_name,
            ExtraArgs=args,
            Config=config,
        )

    def upload_fileobj(
        self,
        file_obj,
        bucket_name,
        key_name,
        storage_class=S3StorageClass.STANDARD,
        extra_args=None,
    ):
        """
        Upload a given fileobj to a specified s3 bucket

        @param file_obj:
        @param bucket_name: target s3 bucket
        @param key_name: The name of the key to upload to
        @param storage_class: the storage class (from Enum) to use for upload.
        @param extra_args: specific args dict to replace the default one provided in this client.
        """
        args = {"ServerSideEncryption": "AES256", "StorageClass": storage_class.name}
        file_obj.seek(0)
        self.s3_client.upload_fileobj(
            Fileobj=file_obj,
            Bucket=bucket_name,
            Key=key_name,
            ExtraArgs=extra_args or args,
        )

    def upload_fileobj_with_timestamp(
        self, file_obj, bucket_name, storage_class=S3StorageClass.STANDARD, prefix=""
    ):
        """
        Upload a given fileobj to a specified s3 bucket with a timestamp formatted s3_key
        @param file_obj:
        @param bucket_name: target s3 bucket
        @param storage_class: the storage class (from Enum) to use for upload.
        @param prefix: prefix to key if any, default to empty
        """

        date_time = datetime.datetime.utcnow().strftime("%Y/%m/%d/%H/%M/%S-%f")[:-4]
        uuid = str(uuid4())
        key_name = "{0}{1}_{2}.gz".format(prefix, date_time, uuid)
        args = {"ServerSideEncryption": "AES256", "StorageClass": storage_class.name}
        self.s3_client.upload_fileobj(
            Fileobj=file_obj, Bucket=bucket_name, Key=key_name, ExtraArgs=args
        )

    def download_file(self, file_name, bucket_name, key_name):
        """
        Download an S3 object to a file

        @param file_name:
        @param bucket_name: target s3 bucket
        @param key_name: The name of the key to download from
        """
        self.s3_client.download_file(
            Bucket=bucket_name, Key=key_name, Filename=file_name
        )

    def download_fileobj(self, file_obj, bucket_name, key_name):
        """
        Download an object from S3 to a file-like object

        @param file_obj: A file-like object to download into
        @param bucket_name:  The name of the bucket to download from
        @param key_name: The name of the key to download from
        """
        self.s3_client.download_fileobj(
            Bucket=bucket_name, Key=key_name, Fileobj=file_obj
        )
        file_obj.seek(0)

    def delete_files(self, bucket_name, key_names):
        """
        Deletes multiple files at provided keys

        @param bucket_name: The name of the bucket to delete from
        @param key_names: List of keys
        """
        if key_names:
            delete_arg = {"Objects": [{"Key": key} for key in key_names]}
            self.s3_client.delete_objects(Bucket=bucket_name, Delete=delete_arg)

    def generate_presigned_download_url(
        self, bucket_name, key_name, expiration_time=3600
    ):
        """
        Generates a presigned URL for an s3 object. Valid for <expiration_date> seconds.
        @param bucket_name: the bucket in which the object is stored
        @param key_name: the key of the object inside the bucket
        @param expiration_time: the amount of time for which the link is valid.
        @return:
        """
        return self.s3_client.generate_presigned_url(
            "get_object",
            Params={"Bucket": bucket_name, "Key": key_name},
            ExpiresIn=expiration_time,
        )

    @staticmethod
    def generate_s3_key_with_timestamp(file_content):
        """
        Generates an s3 key from the hash of the given file with the timestamp

        @param file_content:
        @return: The s3 key with timestamp
        """
        return (
            hashlib.sha256(file_content).hexdigest()
            + "--"
            + datetime.datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%SZ")
        )

    def get_object_summary(self, bucket, key):
        return self.s3_resource.ObjectSummary(bucket, key)

    def get_object_size(self, bucket, key):
        obj_summary = self.get_object_summary(bucket, key)
        return obj_summary.size

    def head_object(self, bucket_name, key_name):
        """
        Returns None if the object doesn't exist.
        """

        try:
            resp = self.s3_client.head_object(Bucket=bucket_name, Key=key_name)
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return None
            else:
                raise e

        return resp

    def store_encrypted_file_with_boto(
        self,
        ciphertext_blob,
        new_iv,
        encrypt_ctx,
        encrypted_file,
        unencrypted_file_size,
        bucket_name,
        key_name,
    ):
        matdesc_string = json.dumps(encrypt_ctx, ensure_ascii=False)
        metadata = {
            "x-amz-key-v2": base64.b64encode(ciphertext_blob).decode("utf-8"),
            "x-amz-iv": base64.b64encode(new_iv).decode("utf-8"),
            "x-amz-cek-alg": "AES/CBC/PKCS5Padding",
            "x-amz-wrap-alg": "kms",
            "x-amz-matdesc": matdesc_string,
            "x-amz-unencrypted-content-length": str(unencrypted_file_size),
        }

        self.s3_client.upload_fileobj(
            Fileobj=encrypted_file,
            Bucket=bucket_name,
            Key=key_name,
            ExtraArgs={"Metadata": metadata},
        )

    def start_list_objects(self, bucket_name, prefix=None):
        self.list_objects_enable = True
        kwargs = {"Bucket": bucket_name, "MaxKeys": 1000}
        if prefix is not None:
            kwargs["Prefix"] = prefix
        while self.list_objects_enable:
            r = self.s3_client.list_objects_v2(**kwargs)
            if "Contents" in r:
                for obj in r["Contents"]:
                    yield obj

                token = r.get("NextContinuationToken")
                if token is None:
                    break
                kwargs["ContinuationToken"] = token
            else:
                break

    def stop_list_objects(self):
        self.list_objects_enable = False

    def get_object_metadata(self, bucket_name, key_name):
        """
        Returns the metadata tags on an object

        @param bucket_name: The name of the bucket to delete from
        @param key_name: The name of the key to delete
        """
        return self.s3_client.head_object(Bucket=bucket_name, Key=key_name)["Metadata"]

    def create_bucket(self, bucket_name):
        """
        Creates a bucket
        @param bucket_name: The name of the bucket to be created
        @return:
        """
        return self.s3_client.create_bucket(Bucket=bucket_name)

    def delete_bucket(self, bucket_name):
        """
        Deletes a bucket
        @param bucket_name: The name of the bucket to be deleted
        @return:
        """
        return self.s3_client.delete_bucket(Bucket=bucket_name)

    def start_list_versions(self, bucket_name, prefix):
        self.list_versions_enable = True
        kwargs = {"Bucket": bucket_name, "Prefix": prefix, "MaxKeys": 1000}

        while self.list_versions_enable:
            r = self.s3_client.list_object_versions(**kwargs)
            if "Versions" in r:
                for version in r["Versions"]:
                    yield version

                next_version_id = r.get("NextVersionIdMarker")
                next_key = r.get("NextKeyMarker")
                if next_version_id is None:
                    break
                kwargs["VersionIdMarker"] = next_version_id
                kwargs["KeyMarker"] = next_key
            else:
                break

    def download_version_fileobj(self, file_obj, bucket_name, key_name, version_id):
        """

        :param file_obj: A file-like object to download into
        :param bucket_name: The name of the bucket to download from
        :param key_name: The key of the object to download from
        :param version_id: Version of the object in S3
        :return: File-like object of content from S3
        """
        self.s3_client.download_fileobj(
            Bucket=bucket_name,
            Key=key_name,
            Fileobj=file_obj,
            ExtraArgs={"VersionId": version_id},
        )
        file_obj.seek(0)

    def copy_file_between_buckets(
        self, source_bucket, source_key, destination_bucket, destination_key
    ):
        """
        Copy a file from one S3 bucket to another.

        Args:
            source_bucket (str): The name of the source bucket.
            source_key (str): The object key of the source file.
            destination_bucket (str): The name of the destination bucket.
            destination_key (str): The object key for the copied file in the destination bucket.
        """
        # Copy the file from the source bucket to the destination bucket
        self.s3_client.copy_object(
            CopySource={"Bucket": source_bucket, "Key": source_key},
            Bucket=destination_bucket,
            Key=destination_key,
        )
        print(
            f"File copied from '{source_bucket}/{source_key}' to '{destination_bucket}/{destination_key}'"
        )
