from tic.shared.utils.util import message_chunk_batch
from uuid import uuid4


class SqsClient(object):
    def __init__(self, boto_session, boto_config=None, endpoint_url=None):
        """

        @param boto_session: Preexisting boto session
        @param boto_config: optional boto config for the sns client
        """

        self.sqs_client = boto_session.client(
            "sqs", config=boto_config, endpoint_url=endpoint_url
        )

    def send_message(self, queue, message):
        """
        Delivers a message to the specified queue

        @param queue: url of sqs queue
        @param message: message to send
        """
        self.sqs_client.send_message(QueueUrl=queue, MessageBody=message)

    def send_message_batch(self, queue, messages):
        """
        Delivers a list of messages to the specified queue

        @param queue: url of sqs queue
        @param messages: the list of SendMessageBatchRequestEntry items to send
        """

        # Batch based on sqs client send_message_batch limit of 10 records or 256 KB limit
        for batch in message_chunk_batch(
            messages, num_records_per_batch=10, max_message_size=256000
        ):

            entries = []
            for message in batch:
                entries.append({"MessageBody": message, "Id": str(uuid4())})

            self.sqs_client.send_message_batch(QueueUrl=queue, Entries=entries)

    def get_queue_url(self, queue_name):
        """
        @param queue_name: the name of the queue
        @return: the url of the queue
        """
        return self.sqs_client.get_queue_url(QueueName=queue_name)["QueueUrl"]

    def receive_message(
        self, queue, wait_seconds=None, num_messages=1, attributes=None
    ):
        """
        Read messages from an SQS Queue
        @param queue: the url of the queue
        @param wait_seconds: The duration (in seconds) for which the call
            will wait for a message to arrive in the queue before returning.
            If a message is available, the call will return sooner than
            wait_time_seconds.
        @param num_messages: The maximum number of messages to read
        @param attributes: The name(s) of additional message
            attributes to return. The default is to return no additional
            message attributes. Use ``['All']`` or ``['.*']`` to return all
        @return: A list of json objects which should contain a list of messages in the `Messages` key
        """
        if attributes is None:
            attributes = list()
        return self.sqs_client.receive_message(
            QueueUrl=queue,
            WaitTimeSeconds=wait_seconds,
            MaxNumberOfMessages=num_messages,
            MessageAttributeNames=attributes,
        )

    def delete_message(self, queue, receipt_handle):
        """

        @param queue: the url of the queue
        @param receipt_handle: the receipt handle of the message
        @return:
        """
        return self.sqs_client.delete_message(
            QueueUrl=queue, ReceiptHandle=receipt_handle
        )

    def change_message_visibility(self, queue, receipt_handle, timeout):
        """

        @param queue: the url of the queue
        @param receipt_handle: the receipt handle of the message
        @param timeout: the new timeout of the desired message
        @return:
        """
        return self.sqs_client.change_message_visibility(
            QueueUrl=queue, ReceiptHandle=receipt_handle, VisibilityTimeout=timeout
        )
