from datetime import datetime

from ....shared.file_acquisition.dtos.file_acquisition_message import (
    FileAcquisitionMessage,
)
from ....shared.utils.date_util import datetime_to_str, str_to_datetime, utc_now
from ....shared.file_acquisition.models.models import (
    Package,
    WsusItem,
)

datetime_format = "%Y-%m-%d %I:%M:%S %p"


class WsusMessage(FileAcquisitionMessage):
    # Encode (json dumps)
    encode_names = {
        "sha256": "Sha256",
        "sha1": "SHA1",
        "md5": "Md5",
        "archive_path": "ArchivePath",
        "update_type": "UpdateType",
        "local_update_id": "LocaUpdateId",
        "title": "Title",
        "file_name": "FileName",
        "download_url": "DownloadUrl",
        "product_family": "ProductFamily",
        "update_classification": "UpdateClassification",
        "kb_article_id": "KbArticleId",
        "security_bulletin_id": "SecurityBulletinId",
        "more_info_url": "MoreInfoUrl",
        "date_created": "DateCreated",
        "date_modified": "DateModified",
        "message_type": "MessageType",
        "source": "Source",
        "operation": "Operation",
        "status": "Status",
    }
    encode_settings = {
        "date_created": lambda x: datetime_to_str(x, datetime_format),
        "date_modified": lambda x: datetime_to_str(x, datetime_format),
    }

    decode_names = {v: k for k, v in encode_names.items()}

    decode_settings = {
        "date_created": lambda x: utc_now()  # TODO update utc now default
        if x is None
        else str_to_datetime(x, datetime_format),
        "date_modified": lambda x: utc_now()
        if x is None
        else str_to_datetime(x, datetime_format),
        "sha256": lambda x: x.lower(),
        "sha1": lambda x: x.lower(),
        "md5": lambda x: x.lower(),
    }

    def __init__(
        self,
        sha256: str = None,
        sha1: str = None,
        md5: str = None,
        archive_path: str = None,
        update_type: str = None,
        local_update_id: int = None,
        title: str = None,
        file_name: str = None,
        download_url: str = None,
        product_family: str = None,
        update_classification: str = None,
        kb_article_id: str = None,
        security_bulletin_id: str = None,
        more_info_url: str = None,
        date_created: datetime = None,
        date_modified: datetime = None,
        message_type: str = None,
        source: str = None,
        operation: str = None,
        status: str = None,
    ):
        self.sha256 = sha256
        self.sha1 = sha1
        self.md5 = md5
        self.archive_path = archive_path  # TODO handle formatting
        self.update_type = update_type
        self.local_update_id = local_update_id
        self.title = title  # TODO handle formatting
        self.file_name = file_name
        self.download_url = download_url
        self.product_family = product_family
        self.update_classification = update_classification
        self.kb_article_id = kb_article_id
        self.security_bulletin_id = security_bulletin_id
        self.more_info_url = more_info_url
        self.date_created = date_created or utc_now()  # operation date
        self.date_modified = utc_now() or date_modified
        self.message_type = message_type
        self.source = source
        self.operation = operation
        self.status = status

    def get_wsus_item(self):
        return WsusItem(
            sha1=self.sha1.upper(),
            archive_path=self.archive_path,
            update_type=self.update_type,
            local_update_id=self.local_update_id,
            # local_update_guid   # TODO
            title=self.title,
            filename=self.file_name,
            download_url=self.download_url,
            product_family=self.product_family,
            update_classification=self.update_classification,
            kb_article_id=str(self.kb_article_id),
            security_bulletin_id=self.security_bulletin_id,
            more_info_url=self.more_info_url,
            date_created=self.date_created,
            date_modified=self.date_modified,
            source_id=self.get_source_id(self.source),
        )

    def get_package(self):
        return Package(
            name=self.file_name,
            sha256=self.sha256.upper(),
            sha1=self.sha1.upper(),
            md5=self.md5.upper(),
            date_created=self.date_modified,
            date_modified=self.date_modified,
        )
