import os
import zipfile
import json
from datetime import datetime, date, timedelta
import re

import dateutil.utils

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
INVALID = "INVALID"
NOT_SIGNED = "NOT_SIGNED"
VALID = "VALID"
EXPIRED = "EXPIRED"
TODAY = date.today().strftime("%Y-%m-%d")


class BaseInfo:
    pass


class HashInfo(BaseInfo):
    def __init__(self):
        self.md5 = ""
        self.sha256 = ""
        self.sha1 = ""
        self.size = 0
        self.file_name = ""
        self.file_type = ""
        self.tlsh = ""

    def __str__(self):
        return (
            f"Algorithm: SHA256,\n"
            f"Hash={self.sha256},\n"
            f"Path={self.file_name},\n"
            f"Sha1={self.sha1},\n"
            f"Md5={self.md5},\n"
            f"Size={self.size}"
        )


class CertificateInfo(BaseInfo):
    def __init__(self):
        self.name = ""
        self.version = 0
        self.serial_number = ""
        self.issuer = ""
        self.issuer_thumbprint = ""
        self.valid_from = None
        self.valid_to = None
        self.thumbprint = ""
        self.status = ""
        self.scan_date = TODAY

    def __str__(self):
        return (
            f"name={self.name},\n"
            f"version={self.version},\n"
            f"serial_number={self.serial_number},\n"
            f"issuer={self.issuer},\n"
            f"issuer_thumbprint={self.issuer_thumbprint},\n"
            f"valid_from={self.valid_from},\n"
            f"valid_to={self.valid_to},\n"
            f"thumbprint={self.thumbprint},\n"
            f"status={self.status},\n"
            f"scan_date={self.scan_date},\n"
        )

    def unique_parameters(self):
        return self.name, self.serial_number

    def __eq__(self, other):
        return (
            self.unique_parameters() == other.unique_parameters()
            if type(other) is type(self)
            else False
        )

    def __hash__(self):
        return hash(self.unique_parameters())


class SignatureInfo(BaseInfo):
    def __init__(self):
        self.sha256 = ""
        self.file_name = ""
        self.date_signed = ""
        self.thumbprint = ""
        self.status = ""

    def __str__(self):
        return (
            f"file={self.file_name},\n"
            f"sha256={self.sha256},\n"
            f"date_signed={self.date_signed},\n"
            f"thumbprint={self.thumbprint},\n"
            f"status={self.status},\n"
        )


class MetadataInfo(BaseInfo):
    def __init__(self):
        self.sha256 = ""
        self.file_name = ""
        self.file_description = ""
        self.file_type = ""
        self.file_version = ""
        self.file_size = ""
        self.product_name = ""
        self.product_version = ""
        self.copyright_details = ""
        self.date_modified = ""
        self.language = ""
        self.original_filename = ""

    def __str__(self) -> str:
        return (
            f"file={self.file_name},\n"
            f"sha256={self.sha256},\n"
            f"file_description={self.file_description},\n"
            f"file_type={self.file_type},\n"
            f"file_version={self.file_version},\n"
            f"file_size={self.file_size},\n"
            f"product_name={self.product_name},\n"
            f"product_version={self.product_version},\n"
            f"copyright_details={self.copyright_details},\n"
            f"date_modified={self.date_modified},\n"
            f"language={self.language},\n"
            f"original_filename={self.original_filename},\n"
        )


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list([ob.__dict__ for ob in obj])
        if isinstance(obj, list):
            return list([ob.__dict__ for ob in obj])
        if isinstance(obj, BaseInfo):
            return obj.__dict__
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


# noinspection SpellCheckingInspection
class CommonUtil:
    exe_mime_types = [
        "application/ecmascript",
        "application/javascript",
        "application/postscript",
        "application/x-csh",
        "application/x-dosdriver",
        "application/x-dosexec",
        "application/x-elf",
        "application/x-executable",
        "application/x-javascript",
        "application/x-mach-binary",
        "application/x-msdos-program",
        "application/x-msdownload",
        "application/x-sh",
        "application/x-sharedlib",
        "application/x-msi",
        "application/vnd.microsoft.portable-executable",
        "text/vbscript",
        "text/x-shellscript",
        "text/x-script.perl",
        "text/x-script.python",
        "text/x-perl",
        "text/x-python",
        "text/javascript",
        "text/x-script.csh",
        "text/x-c++",
        "text/x-Algol68",
        "text/x-msdos-batch",
    ]

    exe_file_extensions = [
        ".com",
        ".cpl",
        ".dll",
        ".drv",
        ".drv",
        ".exe",
        ".js",
        ".ocx",
        ".ps1",
        ".sct",
        ".scr",
        ".sys",
        ".vb",
        ".vbs",
        ".vbscript",
        ".sh",
    ]

    file_extensions_to_ignore = [
        ".aiff",
        ".bmp",
        ".cat",
        ".clmb",
        ".conf",
        ".css",
        ".dtd",
        ".gif",
        ".gz",
        ".h",
        ".html",
        ".ico",
        ".idl",
        ".inf",
        ".icns",
        ".jpeg",
        ".jpg",
        ".json",
        ".log",
        ".manifest",
        ".mfl",
        ".mof",
        ".mp3",
        ".mp4",
        ".mshc",
        ".mui",
        ".mum",
        ".pdb",
        ".pdf",
        ".pem",
        ".plist",
        ".png",
        ".purged",
        ".resjson",
        ".rtf",
        ".strings",
        ".svg",
        ".tif",
        ".tiff",
        ".tmf",
        ".ttc",
        ".ttf",
        ".txt",
        ".xhtml",
        ".xml",
        ".xsl",
        ".zip",
        ".7z",
        ".pod",
        ".pm",
        ".shader",
        ".sql",
        ".nib",
        ".loctable",
    ]

    @staticmethod
    def convert_from_tick_to_datetime(date_tick_format: str) -> str:
        value = datetime(1970, 1, 1) + timedelta(
            seconds=(
                int(re.search("/Date\\((.*)\\)/", date_tick_format).group(1).strip())
            )
            / 1000
        )
        return value.strftime(DATE_FORMAT)

    @staticmethod
    def get_certificate_status(valid_to: datetime) -> str:
        if valid_to < dateutil.utils.today():
            return EXPIRED
        else:
            return VALID

    @staticmethod
    def zip_file(file_list: [], zip_name: str) -> None:
        with zipfile.ZipFile(zip_name, mode="w") as archive:
            for filename in file_list:
                archive.write(filename)

    @staticmethod
    def save_data_to_file(data, output_file: str) -> None:
        with open(output_file, "w") as outfile:
            json.dump(data, outfile, cls=SetEncoder)

    @staticmethod
    def upload_file_to_s3(
        s3_key_name, file_to_upload, hash_count, boto3_session, bucket_name
    ) -> None:
        zip_path = os.path.join(
            "../../os_file_acquisition/os_file_crawler", file_to_upload
        )
        if "mac" in file_to_upload:
            zip_path = os.path.join("/tmp", file_to_upload)
        if "Linux" in file_to_upload:
            zip_path = os.path.join("/tmp", file_to_upload)
        if "Windows" in file_to_upload:
            zip_path = os.path.join(".", file_to_upload)
        print(
            f"Uploading file {file_to_upload} to bucket {bucket_name} / {s3_key_name} having hash count {hash_count}"
        )

        boto3_session.upload_file(
            zip_path,
            bucket_name,
            s3_key_name,
            ExtraArgs={"Metadata": {"hash_count": str(hash_count)}},
        )
        print(f"File {file_to_upload} uploaded successfully to S3 bucket.")
