import pytz

from tic.shared.file_acquisition.models.models import (
    Package,
    PackageSource,
    PackageOperation,
    Status,
    Source,
    WsusItem,
    DropboxItem,
    Operation,
)

from tic.shared.file_acquisition.repository.status_update_db_base import (
    StatusUpdateDbRepositoryBase,
)


class StatusUpdateDbRepository(StatusUpdateDbRepositoryBase):
    def __init__(
        self,
        logger,
        writer_conn_string,
        reader_conn_string,
        connection_args=None,
        statement_timeout=1000,
    ):
        super(StatusUpdateDbRepository, self).__init__(
            logger,
            writer_conn_string,
            reader_conn_string,
            connection_args,
            statement_timeout,
        )

    def get_package(self, sha256, reader):
        """
        Returns the Package Item object corresponding to the sha256, otherwise returns None
        @param sha256: The file hash
        @param reader: Reader session param, file_acquisition_reader_session instance
        @return: The corresponding Package objects, otherwise None
        """

        package = reader.query(Package).filter_by(sha256=sha256.upper()).first()

        return package

    def get_operation(self, name, reader):
        """
        Returns the Operation object corresponding to the operation name, otherwise returns None
        @param name: The operation name
        @param reader: Reader session param, file_acquisition_reader_session instance
        @return: The corresponding Operation objects, otherwise None
        """

        operation = reader.query(Operation).filter_by(name=name.upper()).first()

        return operation

    def get_status(self, name, reader):
        """
        Returns the Status object corresponding to the operation name, otherwise returns None
        @param name: The status name
        @param reader: Reader session param, file_acquisition_reader_session instance
        @return: The corresponding Status objects, otherwise None
        """

        status = reader.query(Status).filter_by(name=name.upper()).first()

        return status

    def get_source(self, name, reader):
        """
        Returns the Source object corresponding to the operation name, otherwise returns None
        @param name: The source name
        @param reader: Reader session param, file_acquisition_reader_session instance
        @return: The corresponding Source objects, otherwise None
        """

        source = reader.query(Source).filter_by(name=name.upper()).first()

        return source

    def get_dropbox_item(self, sha256, reader):
        """
        Returns the Dropbox Item object corresponding to the sha256, otherwise returns None
        @param sha256: The file hash
        @param reader: Reader session param, file_acquisition_reader_session instance
        @return: The corresponding Dropbox Item objects, otherwise None
        """

        dropbox_item = (
            reader.query(DropboxItem).filter_by(sha256=sha256.upper()).first()
        )

        return dropbox_item

    def get_wsus_item(self, sha1, local_update_id, reader):
        """
        Returns the Wsus Item object corresponding to the sha1, otherwise returns None
        @param sha1: The file hash
        @param local_update_id: The local WSUS update ID
        @param reader: Reader session param, file_acquisition_reader_session instance
        @return: The corresponding WSUS Item objects, otherwise None
        """

        wsus_item = (
            reader.query(WsusItem)
            .filter_by(sha1=sha1.upper())
            .filter_by(local_update_id=local_update_id)
            .first()
        )

        return wsus_item

    def get_package_source(self, package, source_id, reader):
        """
        Returns the package source that links the package to the source
        @param package: The Package Object of the item
        @param source: The Source Object of the item
        @param reader: Reader session param, ubs_reader_session instance
        @return: The package source array.
        """

        package_source = (
            reader.query(PackageSource)
            .filter_by(package_id=package.id)
            .filter_by(source_id=source_id)
        ).first()

        return package_source

    def get_package_operation(self, package, operation_id, reader):
        """
        Returns the Package Operations Object that links the package to the operation
        @param package: The Package Object of the item
        @param reader: Reader session param, ubs_reader_session instance
        @return: The Package Operation Object
        """

        package_operation = (
            reader.query(PackageOperation)
            .filter(PackageOperation.package_id == package.id)
            .filter(PackageOperation.operation_id == operation_id)
        ).first()

        return package_operation

    def put_wsus_item(self, wsus_item, writer):
        """
        Puts a record of a WSUS item
        @param wsus_item: the wsus item
        @param writer: Writer session param, file_acquisition_writer_session instance
        @return:
        """
        existing_wsus_item = self.get_wsus_item(
            wsus_item.sha1.upper(), wsus_item.local_update_id, writer
        )

        if existing_wsus_item is not None:
            existing_wsus_item.archive_path = wsus_item.archive_path
            existing_wsus_item.update_type = wsus_item.update_type
            existing_wsus_item.title = wsus_item.title
            existing_wsus_item.filename = wsus_item.filename
            existing_wsus_item.download_url = wsus_item.download_url
            existing_wsus_item.product_family = wsus_item.product_family
            existing_wsus_item.update_classification = wsus_item.update_classification
            existing_wsus_item.kb_article_id = wsus_item.kb_article_id
            existing_wsus_item.security_bulletin_id = wsus_item.security_bulletin_id
            existing_wsus_item.more_info_url = wsus_item.more_info_url
            existing_wsus_item.date_modified = wsus_item.date_modified
            writer.add(existing_wsus_item)
            return existing_wsus_item

        else:
            wsus_item.sha1 = wsus_item.sha1.upper()
            writer.add(wsus_item)
            wsus_item = self.get_wsus_item(
                wsus_item.sha1.upper(), wsus_item.local_update_id, writer
            )
            return wsus_item

    def put_dropbox_item(self, dropbox_item, writer):
        """
        Puts a record of a Dropbox item
        @param dropbox_item: the dropbox item
        @param writer: Writer session param, file_acquisition_writer_session instance
        @return:
        """
        existing_dropbox_item = self.get_dropbox_item(
            dropbox_item.sha256.upper(), writer
        )
        if existing_dropbox_item is not None:
            existing_dropbox_item.name = dropbox_item.name
            existing_dropbox_item.date_modified = dropbox_item.date_modified
            writer.add(existing_dropbox_item)
            return existing_dropbox_item

        else:
            dropbox_item.sha256 = dropbox_item.sha256.upper()
            writer.add(dropbox_item)
            dropbox_item = self.get_dropbox_item(dropbox_item.sha256.upper(), writer)
            return dropbox_item

    def put_package(self, package, writer):
        """
        Puts a record of a Package Item
        @param package: the package item
        @param writer: Writer session param, file_acquisition_writer_session instance
        @return: the Package
        """
        existing_package = self.get_package(package.sha256, writer)
        if existing_package is not None:
            existing_package.name = package.name
            existing_package.sha1 = package.sha1.upper()
            existing_package.md5 = package.md5.upper()
            existing_package.date_modified = package.date_modified
            writer.add(existing_package)
            return existing_package
        else:
            package.md5 = package.md5.upper()
            package.sha1 = package.sha1.upper()
            package.sha256 = package.sha256.upper()
            writer.add(package)
            package = self.get_package(package.sha256, writer)
            return package

    def put_package_source(self, package, source_id, writer):
        """
        Puts a record of a Package Source Item
        @param package: the package item (must have id set)
        @param source: the source item (must have id set)
        @param writer: Writer session param, file_acquisition_writer_session instance
        @return:
        """
        existing_package_source = self.get_package_source(package, source_id, writer)

        if existing_package_source is None:
            package_source = PackageSource(source_id=source_id, package_id=package.id)
            writer.add(package_source)

    def put_package_operation(
        self, package, operation_id, status_id, operation_date, writer
    ):
        """
        Puts a record of a Package Source Item
        @param package: the package item (must have id set)
        @param operation_id: the operation item id
        @param status_id: the status item id
        @param writer: Writer session param, file_acquisition_writer_session instance
        @return:
        """
        existing_package_operation = self.get_package_operation(
            package, operation_id, writer
        )
        if existing_package_operation:
            # Update the status only if the operation date is newer
            if existing_package_operation.operation_date < pytz.UTC.localize(
                operation_date
            ):
                existing_package_operation.status_id = status_id
                existing_package_operation.date_modified = package.date_modified
                existing_package_operation.operation_date = operation_date
                writer.add(existing_package_operation)
            else:
                self.logger.info(
                    f"Newer Package Operation skipped for package op id: {existing_package_operation.id}"
                )
        else:
            package_operation = PackageOperation(
                operation_id=operation_id,
                package_id=package.id,
                status_id=status_id,
                operation_date=operation_date,
            )
            writer.add(package_operation)
