import datetime
import json
import subprocess

from signify.authenticode import AuthenticodeVerificationResult
from signify.authenticode.signed_pe import SignedPEFile

from tic.shared.file_acquisition.os_util import (
    SignatureInfo,
    CertificateInfo,
    MetadataInfo,
    CommonUtil,
    VALID,
    NOT_SIGNED,
    INVALID,
    DATE_FORMAT,
)


# noinspection PyTypeChecker
class WinOsUtil:
    @staticmethod
    def get_signing_and_cert_info(
        sha256: str, file: str, certs: set[CertificateInfo] = None
    ) -> (SignatureInfo, set[CertificateInfo]):
        try:
            signature_info = SignatureInfo()
            file_qualified_path = '"' + file + '"'  # For file path having space
            cs = subprocess.Popen(
                [
                    "powershell.exe",
                    "Get-AuthenticodeSignature -FilePath",
                    file_qualified_path,
                    " | ConvertTo-Json",
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            (o_info, e_info) = cs.communicate()
            status = cs.wait()
            if status != 0 or o_info is None:
                print(f"Error while getting the certificate for file {file}: {e_info}")
                return None, None
            sigs = o_info.decode("utf-8")
            json_data = json.loads(sigs)
            # File is not signed
            if json_data["Status"] == 2 or json_data["SignerCertificate"] is None:
                return None, None
            certificate_info = CertificateInfo()
            certificate_info.name = json_data["SignerCertificate"]["Subject"]
            certificate_info.issuer = json_data["SignerCertificate"]["Issuer"]
            certificate_info.version = json_data["SignerCertificate"]["Version"]
            certificate_info.thumbprint = json_data["SignerCertificate"]["Thumbprint"]
            certificate_info.valid_from = CommonUtil.convert_from_tick_to_datetime(
                json_data["SignerCertificate"]["NotBefore"]
            )
            certificate_info.valid_to = CommonUtil.convert_from_tick_to_datetime(
                json_data["SignerCertificate"]["NotAfter"]
            )
            certificate_info.serial_number = json_data["SignerCertificate"][
                "SerialNumber"
            ]
            certificate_info.status = (
                CommonUtil.get_certificate_status(
                    datetime.datetime.strptime(certificate_info.valid_to, DATE_FORMAT)
                )
                if json_data["Status"] == 0
                else INVALID
            )

            signature_info.thumbprint = certificate_info.thumbprint

            with open(file, "rb") as f:
                digital_signature = SignedPEFile(f)
                status, err = digital_signature.explain_verify()
                if status == AuthenticodeVerificationResult.OK:
                    signature_info.date_signed = (
                        list(digital_signature.signed_datas)[0]
                        .signer_infos[0]
                        .countersigner.signing_time
                    )
                    signature_info.status = VALID
                elif status == AuthenticodeVerificationResult.NOT_SIGNED:
                    signature_info.status = NOT_SIGNED
                else:
                    signature_info.status = INVALID

            signature_info.file_name = file
            signature_info.sha256 = sha256

            if signature_info.status == NOT_SIGNED or status is None:
                return None, None

            already_exists = False
            if certs is not None:
                for cert in certs:
                    if cert.thumbprint == certificate_info.thumbprint:
                        already_exists = True
                        break
            if already_exists:
                return signature_info, None

            all_certificates = WinOsUtil.get_issuer_certificates(file, certificate_info)
            return signature_info, all_certificates
        except Exception as e:
            print(f"Certificate error: {file} - {e}")
            return None, None

    @staticmethod
    def get_issuer_certificates(file, signer_certificate) -> set[CertificateInfo]:
        all_certificates: set[CertificateInfo] = set()
        certificate_dict: dict[str, CertificateInfo] = {
            signer_certificate.issuer: signer_certificate
        }
        all_certificates.add(signer_certificate)
        file_qualified_path = '"' + file + '"'  # For file path having space
        cs = subprocess.Popen(
            [
                "powershell.exe",
                " $cert = (Get-AuthenticodeSignature -FilePath ",
                file_qualified_path,
                " ).SignerCertificate; ",
                " $chain = New-Object -TypeName System.Security.Cryptography.X509Certificates.X509Chain; "
                " $chain.Build($cert);"
                " $chain.ChainElements | ForEach-Object {$_.Certificate} | ConvertTo-Json",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        (o_info, e_info) = cs.communicate()
        status = cs.wait()
        if status != 0 or o_info is None:
            return set()
        sigs = o_info.decode("utf-8")
        dataform = (
            str(sigs).replace("True", "").replace("False", "")
        )  # Removes first line
        json_data = json.loads(dataform)

        certs = json_data if type(json_data) is list else [json_data]

        for certificate in certs:
            if signer_certificate.thumbprint == certificate["Thumbprint"]:
                continue
            certificate_info = CertificateInfo()
            certificate_info.name = certificate["Subject"]
            certificate_info.issuer = certificate["Issuer"]
            certificate_info.version = certificate["Version"]
            certificate_info.thumbprint = certificate["Thumbprint"]
            certificate_info.valid_from = CommonUtil.convert_from_tick_to_datetime(
                certificate["NotBefore"]
            )
            certificate_info.valid_to = CommonUtil.convert_from_tick_to_datetime(
                certificate["NotAfter"]
            )
            certificate_info.serial_number = certificate["SerialNumber"]
            certificate_info.status = CommonUtil.get_certificate_status(
                datetime.datetime.strptime(certificate_info.valid_to, DATE_FORMAT)
            )
            if (
                certificate_info.issuer == certificate_info.name
            ):  # For Root CA certificate
                certificate_info.issuer_thumbprint = certificate_info.thumbprint

            all_certificates.add(certificate_info)
            # If current certificate CN matches any existing certificate's Issuer CN,
            # Update existing certificate's issuer_thumbprint
            if certificate["Subject"] in certificate_dict:
                temp_cert = certificate_dict.get(certificate["Subject"])
                temp_cert.issuer_thumbprint = certificate["Thumbprint"]
                all_certificates.add(temp_cert)
            certificate_dict[certificate_info.issuer] = certificate_info
        return all_certificates

    @staticmethod
    def get_metadata_info(sha256, file) -> MetadataInfo:
        try:
            temp_file = '"' + file + '"'  # For file path having space
            cs = subprocess.Popen(
                ["powershell.exe", " Get-Item ", temp_file, " | ConvertTo-Json"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            (o_info, e_info) = cs.communicate()
            status = cs.wait()
            if status != 0 or o_info is None:
                return None
            sigs = o_info.decode("utf-8")
            json_data = json.loads(sigs)
            metadata_info = MetadataInfo()
            metadata_info.sha256 = sha256
            metadata_info.file_name = file
            metadata_info.file_description = json_data["VersionInfo"]["FileDescription"]
            metadata_info.file_type = json_data["Extension"]
            metadata_info.file_version = json_data["VersionInfo"]["FileVersion"]
            metadata_info.file_size = json_data["Length"]
            metadata_info.product_name = json_data["VersionInfo"]["ProductName"]
            metadata_info.product_version = json_data["VersionInfo"]["FileVersion"]
            metadata_info.copyright_details = json_data["VersionInfo"]["LegalCopyright"]
            metadata_info.date_modified = CommonUtil.convert_from_tick_to_datetime(
                json_data["LastWriteTime"]
            )
            metadata_info.language = json_data["VersionInfo"]["Language"]
            metadata_info.original_filename = json_data["VersionInfo"][
                "OriginalFilename"
            ]

            return metadata_info
        except Exception as e:
            print(f"Metadata error: {file} -{e}")
            return None
