from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker


class UbsDbRepositoryBase(object):
    def __init__(
        self,
        logger,
        writer_conn_string=None,
        reader_conn_string=None,
        connection_args=None,
        statement_timeout=0,
    ):
        statement_timeout = int(statement_timeout)
        if connection_args is None:
            # This should always be set
            # For local mcs2: sslmode=disable
            # For prod and lambdas: sslmode=verify-ca
            # For CI=prefer # TODO: Can this be made verify-ca?
            connection_args = dict()
        connection_args.setdefault("options", f"-c {statement_timeout=}")
        # Pass in {'sslmode': 'verify-ca', 'sslrootcert': '/tmp/root.crt'} for Lambda.

        self.logger = logger

        if writer_conn_string:
            # Records connections (generic sql only, but returns dicts):
            self.writer_engine = create_engine(
                writer_conn_string, connect_args=connection_args
            )
            self.writer = sessionmaker(bind=self.writer_engine)
        if reader_conn_string:
            self.reader = sessionmaker(
                bind=create_engine(reader_conn_string, connect_args=connection_args),
                autoflush=False,
                autocommit=False,
            )

    @contextmanager
    def ubs_reader_session(self):
        """
        Provide a transactional scope around a reader instance.
        """
        session = scoped_session(self.reader)
        try:
            yield session
        finally:
            session.close()

    @contextmanager
    def ubs_writer_session(self):
        """
        Provide a transactional scope around a writer instance.
        """
        session = scoped_session(self.writer)
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
