from datetime import datetime
from time import mktime, gmtime

# The beginning of time for epoch time
EPOCH = datetime(1970, 1, 1)


def to_iso_8601(dt, fractional_seconds=True):
    if fractional_seconds:
        return datetime_to_str(dt, "%Y-%m-%dT%H:%M:%S.%fZ")
    return datetime_to_str(dt, "%Y-%m-%dT%H:%M:%SZ")


def from_iso_8601(dt, fractional_seconds=True):
    if fractional_seconds:
        return str_to_datetime(dt, "%Y-%m-%dT%H:%M:%S.%fZ")
    return str_to_datetime(dt, "%Y-%m-%dT%H:%M:%SZ")


def match_str_strptime_to_datetime(s):
    """
    Match an expected ISO 8601 string to a function to return a datetime object
    """

    parts = s.split(".")

    if len(parts) == 1:
        fmt = "%Y-%m-%dT%H:%M:%SZ"
    else:
        # :-1 Chops off the Z. :26 takes in the entire timestamp up to microseconds.
        s = s[:-1][:26] + "Z"
        fmt = "%Y-%m-%dT%H:%M:%S.%fZ"

    return str_to_datetime(s, fmt)


def utc_now():
    """
    Get a datetime object representing now in UTC.
    @return: A datetime object for UTC now.
    """

    return datetime.utcnow()


def utc_datetime_to_epoch_time(dt):
    """
    Convert a datetime object to epoch time.
    It is assumed that the datetime object is in epoch time.

    @param dt: A datetime object that is already in UTC.
    @return: Epoch time.
    """

    return (dt - EPOCH).total_seconds()


def epoch_time_to_utc_datetime(x):
    """
    Convert an epoch time to a datetime object.

    @param x: The epoch time.
    @return: A datetime object that is in UTC.
    """

    return datetime.utcfromtimestamp(x)


def str_to_datetime(s, fmt):
    """
    Convert a string, with the specified format to a datetime object.

    @param s: The string to convert.
    @param fmt: The format of the string.
    @return: A datetime object.
    """
    return datetime.strptime(s, fmt)


def str_to_datetime_chop_nanoseconds(s, fmt):
    """
    Convert a string, with the specified format to a datetime object
    while chopping off the nanoseconds.

    @param s: The string to convert - assumes nanoseconds in the string and it ends with a 'Z'
    @param fmt: The format of the string.
    @return: A datetime object.
    """

    return str_to_datetime(s[:-4] + "Z", fmt)


def datetime_to_str(dt, fmt):
    """
    Convert a datetime object to a string with the specified format.

    @param dt: The datetime object.
    @param fmt: The format of the string.
    @return: The formatted string.
    """
    if dt is None:
        return utc_now().strftime(fmt)
    elif dt.year < 1970:
        return datetime.utcfromtimestamp(mktime(gmtime(0))).strftime(fmt)
    else:
        return dt.strftime(fmt)


def datetime_default(dt):
    """
    Returns the datetime object if it is not None; otherwise, returns a datetime object
    representing UTC now.

    @param dt: The datetime object to use.
    @return: A new datetime object.
    """
    if dt is None:
        return utc_now()
    elif dt.year < 1970:
        return datetime.utcfromtimestamp(mktime(gmtime(0)))
    else:
        return dt
