from tic.shared.utils.parsers import network
from .validation_base import Validation, pass_score_threshold


class ValidateDNS(Validation):
    def __init__(
        self,
        input_data=None,
        validation_types=("dns_cleanlist",),
        score_threshold=0,
        **kwargs
    ):
        """
        :param validation_types: Initialize with this validation type.
        :param score_threshold: Not really used here yet - just True/False Passing
        :param kwargs: Other arguments passed to ValidationBase
        :return:
        """
        if input_data is None:
            input_data = list()
        super(ValidateDNS, self).__init__(**kwargs)

        self._ioc_type = "dns"

        self.set_input_data(input_data)

        for validation_type in validation_types:
            if validation_type == "dns_cleanlist":
                self.add_validation_method("dns_cleanlist", check_dns_cleanlist)
                self.add_score_threshold_method(
                    "dns_cleanlist", pass_score_threshold, score_threshold
                )
            else:
                raise ValueError("Invalid validation type {0}".format(validation_type))

    def set_input_data(self, input_dns):
        """
        Will raise an Exception if any of the elements are invalid when adding the input data.

        :param input_dns: List of Domain Names.
        :return: None - Will set an internal variable of bad_iocs.
        """

        for d in input_dns:
            network_type, temp_network = network.determine_authority_type(d)
            if network_type != network.AuthorityType.domain:
                self.bad_input_data.append(d)
            else:
                self._input_data.append(d)

        if self.bad_input_data:
            raise ValueError("ValidateDNS: Invalid IOCs Inputted")


def check_dns_cleanlist(input_domain, results):
    """
    This should eventually become a database and cleanlist to check - with the domains as TLDs we should trust.

    :param input_domain: List of Domains to check
    :return:
    """
    try:
        for domain in input_domain:
            domain_lowered = domain.lower()
            right_2_labels = ".".join(domain_lowered.split(".")[-2:])
            right_3_labels = ".".join(domain_lowered.split(".")[-3:])

            if right_2_labels in DNS_CLEANLIST:
                continue
            elif right_3_labels in DNS_CLEANLIST:
                continue
            else:
                results.append(domain)

        return results
    except Exception as e:
        # Delete all elements in the list
        del results[:]

        # Set the list to one element that is the exception for this thread.
        results.append(e)
        return None


# All of these should be lowercase.
DNS_CLEANLIST = {
    # Alexa Top 100
    "google.com",
    "facebook.com",
    "youtube.com",
    "baidu.com",
    "yahoo.com",
    "amazon.com",
    "wikipedia.org",
    "qq.com",
    "google.co.in",
    "twitter.com",
    "live.com",
    "taobao.com",
    "msn.com",
    "sina.com.cn",
    "yahoo.co.jp",
    "linkedin.com",
    "google.co.jp",
    "weibo.com",
    "bing.com",
    "vk.com",
    "yandex.ru",
    "hao123.com",
    "ebay.com",
    "instagram.com",
    "google.de",
    "amazon.co.jp",
    "mail.ru",
    "360.cn",
    "google.ru",
    "google.co.uk",
    "pinterest.com",
    "t.co",
    "google.com.br",
    "reddit.com",
    "netflix.com",
    "google.fr",
    "tmall.com",
    "sohu.com",
    "paypal.com",
    "microsoft.com",
    "wordpress.com",
    "blogspot.com",
    "google.it",
    "tumblr.com",
    "onclickads.net",
    "google.es",
    "imgur.com",
    "imdb.com",
    "aliexpress.com",
    "ok.ru",
    "xvideos.com",
    "apple.com",
    "ask.com",
    "fc2.com",
    "gmw.cn",
    "stackoverflow.com",
    "google.com.mx",
    "amazon.de",
    "google.com.hk",
    "alibaba.com",
    "google.ca",
    "google.com.tr",
    "office.com",
    "rakuten.co.jp",
    "pornhub.com",
    "xinhuanet.com",
    "soso.com",
    "google.co.id",
    "tianya.cn",
    "amazon.in",
    "craigslist.org",
    "haosou.com",
    "blogger.com",
    "github.com",
    "amazon.co.uk",
    "nicovideo.jp",
    "kat.cr",
    "outbrain.com",
    "bongacams.com",
    "adnetworkperformance.com",
    "go.com",
    "360.com",
    "google.pl",
    "googleusercontent.com",
    "cnn.com",
    "naver.com",
    "pixnet.net",
    "google.com.au",
    "dropbox.com",
    "flipkart.com",
    "adobe.com",
    "xhamster.com",
    "jd.com",
    "chinadaily.com.cn",
    "whatsapp.com",
    "microsoftonline.com",
    "chase.com",
    "bbc.co.uk",
    "coccoc.com",
    "youku.com",
    # Manually added - 2016-02-23 - RJPECK
    "carbonblack.com",
    "carbonblack.io",
    "bit9.com",
    # TPLAT-1007 Request
    "goo.gl",
    "hybrid-analysis.com",
    "malshare.com",
}
