"""
import gevent.monkey
gevent.monkey.patch_all()
Validation functionality for hashes types (md5 only right now). Including functions to query internal TIC services.

This is intended to validate lists of IOCs.

Can receive a logging object and will use. Otherwise will set up a standard
logger to go to STDERR/STDOUT. Pass skip to have no logs.

"""
# TODO: temporary to fix
try:
    from tic.shared.utils.discovery_util import get_healthy_service_uri
    from requests_futures.sessions import FuturesSession
    from requests.exceptions import ReadTimeout
except ImportError:
    pass

from tic.shared.utils.parsers import hash as parser_hash
from tic.shared.utils.util import chunk

from .validation_base import Validation


class ValidateHash(Validation):
    # List of valid hash types that all of the input hashes should be.
    _valid_hash_types = [
        parser_hash.HashType.md5,
        parser_hash.HashType.sha1,
        parser_hash.HashType.sha256,
    ]

    def __init__(
        self,
        input_data=None,
        validation_types=("srs",),
        score_threshold=-50,
        pass_through_header=None,
        **kwargs,
    ):
        """
        :param input_data: Input IOCs to Validate
        :param validation_types: Initialize with this validation type (srs only right now).
        :param score_threshold: Initialize with this score threshold (defaults to -50)
        :param kwargs: Other arguments passed to ValidationBase
        :return:
        """
        if input_data is None:
            input_data = list()
        super(ValidateHash, self).__init__(**kwargs)

        self.pass_through_header = pass_through_header
        self._ioc_type = "hash"

        self._validation_score_ranges = {"srs": (-100, 100)}

        self.set_input_data(input_data)

        extra_args = {}
        if self.pass_through_header is not None:
            extra_args["input_headers"] = self.pass_through_header

        for validation_type in validation_types:
            if validation_type == "srs":
                self.add_validation_method(
                    "srs", query_srs_internal, extra_args=extra_args
                )
                self.add_score_threshold_method(
                    "srs", srs_score_threshold, score_threshold
                )

    def set_input_data(self, input_hashes):
        """
        Will raise an Exception if any of the hashes are invalid when adding the input data.

        :param input_hashes: List of hashes that are of mixed types - must be MD5, SHA1 or SHA256.
        :return: Will set an internal variable of bad_iocs.
        """
        for h in input_hashes:
            hash_type, temp_hash = parser_hash.determine_hash_type(h)
            if hash_type not in self._valid_hash_types:
                self.bad_input_data.append(h)
            else:
                self._input_data.append(temp_hash.lower())

        if self.bad_input_data:
            raise ValueError("ValidateHash: Invalid IOCs Inputted")


def srs_score_threshold(srs_results, score_threshold, include_none=True):
    """

    :param srs_results: List of the results from SRS
    :param score_threshold: The Score Threshold (normalized score -100 to 100)
    :param include_none: if True - will return hashes that didn't have a score - else will not.
    :return: yield the hashes that are above the threshold
    """

    for result in srs_results:
        parts = result["reputation"]["score"]["parts"]
        reputation = parts[0] if len(parts) > 0 else None
        if (
            reputation is None
            or reputation["score"] is None
            or reputation["score"] == 0
        ) and include_none:
            yield result["hash"].lower()

        # Comparing something to None won't return true.
        if (
            reputation
            and reputation.get("score")
            and reputation["score"] >= score_threshold
        ):
            yield result["hash"].lower()


def query_srs_internal(
    hash_list, results, input_headers=None, batch_size=100, timeout=30
):
    """
    Query the hash reputation service for a list of hashes

    @param hash_list:
    @param input_headers: any auth headers you need.
    @param batch_size: Number of hashes to send at a time to hash_reputation.
    @param timeout: the timeout of each parallel hash reputation call.
    @return Dictionary of the hashes and the scores. None if not found.
    """
    if input_headers is None:
        input_headers = dict()
    try:
        headers = {"Content-Type": "application/json"}
        headers.update(input_headers)

        req_session = FuturesSession()
        service_uri = get_healthy_service_uri("hashreputation", "v1", "hash")

        async_requests = [
            req_session.post(
                service_uri, json={"hashes": batch}, timeout=timeout, headers=headers
            )
            for batch in chunk(hash_list, batch_size)
        ]

        # Merge the response and extract the relevant data
        for req in async_requests:
            try:
                response = req.result()
                assert response.status_code == 200
            except (ReadTimeout, AssertionError) as e:
                error_message = (
                    "Call in query_srs_internal failed to receive any response."
                )
                if isinstance(e, AssertionError):
                    error_message += "url = {}, status_code = {}".format(
                        response.url, response.status_code
                    )
                raise ValueError(error_message)

            results.extend(response.json()["data"])

        return results
    except Exception as e:
        # Delete all elements in the list
        del results[:]

        # Set the list to one element that is the exception for this thread.
        results.append(e)
        return None
