"""
    Shared utilities for Kinesis Lambda handlers
"""
import json
import traceback
from sqlalchemy.exc import OperationalError
from sqlalchemy.exc import DataError
from botocore.exceptions import ConnectionClosedError
from botocore.exceptions import ClientError

from tic.shared.utils.date_util import utc_now
from tic.shared.utils.exception_util import LambdaException


class KinesisLambdaUtil(object):
    def __init__(self, in_dlq_name, in_sqs_client, in_logger, in_statsd):
        self.dlq_name = in_dlq_name
        self.sqs_client = in_sqs_client
        self.logger = in_logger
        self.statsd = in_statsd

    @staticmethod
    def create_dlq_records(ex, in_records, record_key):
        """
        create records to put into a dlq
        @param ex: exception instance
        @param in_records: records to be processed to put to a dlq
        @param record_key: "raw_record" - non-replayable records, "record" - replayble records
        @return: list of records ready to put to a dle
        """
        batch = []
        ex_msg = str(ex)
        st = traceback.format_exc()
        for record in in_records:
            batch.append(
                {"exception_message": ex_msg, "stack_trace": st, record_key: record}
            )
        return batch

    @staticmethod
    def create_message(dlq_records):
        """
        create list individual message not chunked before putting to a dlq
        @param dlq_records: records to be put to a dlq
        @return: list of individual record to be put to dlq
        """
        return [
            json.dumps(x, separators=(",", ":"), ensure_ascii=False)
            for x in dlq_records
        ]

    @classmethod
    def create_dlq_messages(cls, ex, in_records, in_key="bad_record"):
        """
        create records ready to put to a dlq
        @param ex: exception instance
        @param in_records: records to be put to a dlq
        @param in_key:
            "raw-record": non-replayable raw-record,
            "bad-record": non-replayable json loaded record
            "record": maybe replayable json loaded record
        @return: list of not chunked records ready to put a dlq
        """
        batch = cls.create_dlq_records(ex, in_records, in_key)
        return cls.create_message(batch)

    def submit_to_dlq(self, messages):
        """
        submit dlq ready formatted messages to the sqs client belonging to this handler
        @param messages: dlq readily formatted messages
        @return: None
        """
        if len(messages) == 0:
            self.logger.warning("Nothing to put to DLQ, messages queue is empty")

        self.sqs_client.send_message_batch(
            queue=self.sqs_client.get_queue_url(queue_name=self.dlq_name),
            messages=messages,
        )

    def send_lambda_exception_to_dlq(self, in_lambda_e_data):
        """
        send records in LambdaException to dlq
        @param in_lambda_e_data: data in LambdaException instance
        @return: None
        """
        message = self.create_message(in_lambda_e_data)
        self.submit_to_dlq(message)

    def get_timing_stats(self, start_time):
        """
        get timing stats from lambda handler
        @param start_time:
        @return:
        """
        self.statsd.timing(
            "lambda.handler.uncaught.exception.processing.time",
            utc_now() - start_time,
        )

    def increase_retryable_counters(self, in_records=None):
        if in_records is None:
            return

        self.statsd.incr("lambda.handler.exception.retryable.invocation", 1)
        self.statsd.incr("lambda.handler.exception.retryable.records", len(in_records))

    def exception_handler(self, ex, in_raw_records=None, in_records=None):
        """
        exception handler
        @param ex: exception instance
        @param in_raw_records: non-replayable records
        @param in_records: replayable records
        @return: None
        """
        start_time = utc_now()
        if in_raw_records is not None and in_records is not None:
            self.logger.exception(
                "Bad input to exception_handler: both raw and parsed records were provided "
            )
            self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
            self.statsd.incr("lambda.handler.exception.non_retryable.records", 1)
            raise ex

        if in_raw_records is None and in_records is None:
            self.logger.exception("nonexistent raw_records and records")
            self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
            self.statsd.incr("lambda.handler.exception.non_retryable.records", 1)
            raise ex

        if in_raw_records is not None:
            # raw_record failed to be json loaded/decoded, dump to DLQ
            self.logger.exception(
                "Exception caught when json decoding/loading raw records."
            )
            messages = self.create_dlq_messages(ex, in_raw_records, "raw_record")
            self.submit_to_dlq(messages)
            self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
            self.statsd.incr(
                "lambda.handler.exception.non_retryable.records", len(in_raw_records)
            )
            self.get_timing_stats(start_time)
        else:
            if isinstance(ex, OperationalError):
                self.logger.exception("Rethrow OperationalError")
                self.increase_retryable_counters(in_records)
                self.get_timing_stats(start_time)
                raise ex
            elif isinstance(ex, LambdaException):
                self.logger.exception("Rethrow retryable LambdaException")
                self.increase_retryable_counters(in_records)
                self.get_timing_stats(start_time)
                raise ex
            elif isinstance(ex, ConnectionClosedError):
                # this exception was observed when ubs was scaled down, which in theory should be a temporary situation
                self.logger.exception(
                    "Connection was closed before a valid response from endpoint was received - rethrow exception"
                )
                self.increase_retryable_counters(in_records)
                self.get_timing_stats(start_time)
                raise ex
            elif isinstance(ex, ClientError):
                # this exception was observed when ubs / kinesis were scaled down, which in theory should be a transient situation
                self.logger.exception(
                    "ClientError: InternalFailure when calling the PutRecords operation - rethrow exception"
                )
                self.increase_retryable_counters(in_records)
                self.get_timing_stats(start_time)
                raise ex
            elif isinstance(ex, DataError):
                self.logger.exception("DataError: flagged as bad_record")
                messages = self.create_dlq_messages(ex, in_records, "bad_record")
                self.submit_to_dlq(messages)
                self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
                self.statsd.incr(
                    "lambda.handler.exception.non_retryable.records", len(in_records)
                )
                self.get_timing_stats(start_time)
            else:
                self.logger.exception("Unexpected exception caught")
                # may or may not be replayable, keyed with "record"
                messages = self.create_dlq_messages(ex, in_records, "record")
                self.submit_to_dlq(messages)
                self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
                self.statsd.incr(
                    "lambda.handler.exception.non_retryable.records", len(in_records)
                )
                self.get_timing_stats(start_time)
