import json


def chunk(iterable, chunk_size):
    """
    Yields chunks of the specified size from the provided iterable.

    @param iterable: The item to iterate across.
    @param chunk_size: The size of the chunk to yield.
    @return: The chunks.
    """

    items = []

    for item in iterable:
        # Add elements until batch size is reached
        items.append(item)
        if len(items) != chunk_size:
            continue

        # Yield the item and reset the list
        yield items
        items = []

    # Yield any remaining items
    if len(items) > 0:
        yield items


def message_chunk_batch(
    iterable, num_records_per_batch=None, max_message_size=None, delimiter=""
):

    """
    Yields correct message batch that is equal to the max records limit or the max message size
    @param iterable: The record with a 'Data' field to be split
    @param num_records_per_batch: The number of records per batch
    @param max_message_size: The max message size to be sent in a single batch
    @return: Batches of the specified record size or of the max message size
    """

    length_counter = 0
    batched_records = []

    for message in iterable:
        # Handle two main types we expect to be passed to this function.
        if isinstance(message, dict):
            temp_message = json.dumps(message, ensure_ascii=False)

            if delimiter:
                raise ValueError("delimiter must be empty when passing dict messages")
        elif isinstance(message, bytes):
            message = message.decode("utf-8") + delimiter
            temp_message = message + delimiter
        else:
            message = message + delimiter
            temp_message = message + delimiter

        # Update the length counter
        length_counter += len(temp_message)

        if (
            num_records_per_batch is not None
            and len(batched_records) >= num_records_per_batch
        ) or (max_message_size is not None and length_counter > max_message_size):

            yield batched_records
            length_counter = len(temp_message)
            batched_records = []

        batched_records.append(message)

    # Yield any remaining items
    if len(batched_records) > 0:
        yield batched_records


def balanced_split(iterable, desired_chunk_number):
    """
    Yields a specified number of balanced chunks of an iterable
    @param iterable: the iterable to split
    @param desired_chunk_number: The desired number of chunks.
    @return: Chunks
    """
    iterable = list(iterable)
    base_divisor = len(iterable) // desired_chunk_number
    rem = len(iterable) % desired_chunk_number

    count = 0
    while count < len(iterable):
        to_yield = iterable[count : (count + base_divisor)]
        count += base_divisor
        if rem > 0:
            to_yield.append(iterable[count])
            rem -= 1
            count += 1
        yield to_yield


def balanced_chunk(iterable, desired_chunk_size, allow_size_increase=True):
    """
    Yields chunks of evenly distributed sizes.
    CAUTION: This method requires that the iterable be loaded into memory.

    @param iterable: The iterable to utilize.
    @param desired_chunk_size: The desired chunk size to use.
    @param allow_size_increase: If True, non-evenly divisible chunks sizes will
    result in some batches being larger than the desired_chunk_size. If False,
    non-evenly divisible chunks sizes will result in some batches being smaller
    than the desired_chunk_size. In the former, there will be
    len(iterable) / desired_chunk_size iterations. In the latter, there will be
    len(iterable) / (desired_chunk_size - 1) iterations.
    @return: The chunks.
    """

    items = list(iterable)
    n_items = len(items)

    # If the iterable is empty - do nothing
    if n_items == 0:
        return

    remainder = n_items % desired_chunk_size

    def remainder_chunk(_items, _chunk_size, _remainder):
        """
        Yield chunks of data, evenly, adding the remaining items into the first chunks.
        """

        data = []
        for item in _items:
            # Add items while the desired amount is not reached
            data.append(item)
            if len(data) < _chunk_size:
                continue

            # Add an additional item if there is a remainder
            if len(data) == _chunk_size:
                if _remainder > 0:
                    _remainder -= 1
                    continue

            # Yield the data and reset the list
            yield data
            data = []

    # Check for small batch
    if n_items < desired_chunk_size:
        generator = [items]
    # Check for equal items
    elif remainder == 0:
        generator = chunk(items, desired_chunk_size)
    # Max batch size is not enforced, so increase batch size
    elif allow_size_increase:
        generator = remainder_chunk(items, desired_chunk_size, remainder)
    # Max batch size is enforced, so decrease batch size
    else:
        chunk_size = desired_chunk_size - 1
        remainder = n_items % chunk_size
        generator = remainder_chunk(items, chunk_size, remainder)

    # Yield the items (in Py3 can use yield from)
    for item in generator:
        yield item


def replace_dict_value(d, keys, func):
    """
    Replace the value in the dictionary at the specified location with the desired value.
    This method will only replace the value if the keys exists.

    @param d: The dictionary to manipulate.
    @param keys: An iterable containing the keys directing to the value to be replaced.
    @param func: Lambda function defining how to replace the value. The input to the lambda
    will be the old value.
    """

    c_d = p_d = d
    obj = object()
    k = None
    for k in keys:
        if c_d is None or c_d is obj:
            return
        p_d, c_d = c_d, c_d.get(k, obj)
    if k is not None and c_d is not obj:
        p_d[k] = func(c_d)


def to_bool(param):
    return str(param).lower() in {"true", "1"}
